import numpy as np
import sys
import os
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from ttgc import helpers
from ttgc.globals import can_use_GPU

if can_use_GPU:    
    import torch
    # Helper function to convert between numpy arrays and tensors
    to_t = lambda array: torch.tensor(array, device=device, dtype=dtype)
    from_t = lambda tensor: tensor.to("cpu").detach().numpy()
    device = torch.device('cuda')  # Use GPU
    dtype = torch.float32

def calc_square_dist_squared(mat, use_GPU):
    """
    Calculates the square of the square norm (to generate a non-twisted torus).
    
    Args:
        mat (np.ndarray): array of shape (2, N, N) containing the differences in positions between pairs of cells, either with or without velocity modulation.
        use_GPU (bool): whether to use GPU.
    Returns:
        square_dist_squared (np.ndarray): array of shape (2, N, N) containing the square_norms on the modulated position differences array.
    """
    n_comps = mat.shape[1]
    S = np.array([[0, 0], [1, 1], [0, 1], [1, 0],
                  [1, -1], [-1, 1], [-1, -1], [0, -1], [-1, 0]])
    S = np.fliplr(S)
    
    # make copies to do vector subtractions
    S = np.swapaxes(np.expand_dims(S, 1), 2, 0)
    S = np.tile(S, (1, n_comps, 1))

    mat = np.expand_dims(mat, 2)
    mat = np.tile(mat, (1, 1, 9))
    sum_mat = S + mat
    
    if (use_GPU):
        t_sum_mat = to_t(sum_mat)
        t_square_dist_squared = torch.min(torch.sum(torch.pow(t_sum_mat, 2), 0), 1)[0]
        square_dist_squared = from_t(t_square_dist_squared.cpu())
    else:
        square_dist_squared = np.min(np.sum(np.power(sum_mat, 2), axis=0), axis=1)

    return square_dist_squared

def calc_tri_dist_squared(mat, use_GPU):
    """
    Calculates the square of the tri-norm as defined in Guanella et al. 2007 (see equations 4-13, and the numerator of equation 14).
    
    Args:
        mat (np.ndarray): array of shape (2, N^2) containing the differences in positions between pairs of cells, either with or without velocity modulation.
        use_GPU (bool): whether to use GPU.
    Returns:
        tri_dist_squared (np.ndarray): array of shape (2, N, N) containing the tri-norms on the modulated position differences array.
    """
    n_comps = mat.shape[1]
    S = np.array([[0, 0], [-0.5, np.sqrt(3)/2], [-0.5, -np.sqrt(3)/2], 
                  [0.5, np.sqrt(3)/2], [0.5, -np.sqrt(3)/2], [-1, 0], [1, 0]])
    S = np.fliplr(S)
    
    # make copies to do vector subtractions
    S = np.swapaxes(np.expand_dims(S, 1), 2, 0)
    S = np.tile(S, (1, n_comps, 1))
    
    mat = np.expand_dims(mat, 2)
    mat = np.tile(mat, (1, 1, 7))
    sum_mat = S + mat

    if (can_use_GPU and use_GPU):
        t_sum_mat = to_t(sum_mat)
        t_tri_dist_squared = torch.min(torch.sum(torch.pow(t_sum_mat, 2), 0), 1)[0]
        tri_dist_squared = from_t(t_tri_dist_squared.cpu())
    else:
        tri_dist_squared = np.min(np.sum(np.power(sum_mat, 2), axis=0), axis=1)

    return tri_dist_squared

def calc_weight_mat(cell_position_diffs, n_y, n_x, alpha, beta, velocity, I, sigma, T, use_GPU):
    """
    Calculates the weight matrix used to calculate the contributions of each grid cell to the linear transfer function.
    
    Args:
        cell_position_diffs (np.ndarray): array of shape (2, N, N), where N = n_y * n_x, containing the differences in positions between pairs of cells. 
        n_y (int): number of grid cells along the y-axis.
        n_x (int): number of grid cells along the x_axis.
        alpha (float): gain.
        beta (float): bias.
        velocity (np.ndarray): array of shape (2,) representing the velocity at a timebin.
        I (float): intensity parameter.
        sigma (float): Gaussian size.
        T (float): shift parameter.
        use_GPU (bool): whether to use GPU.
    Returns:
        W (np.ndarray): array of shape (N, N) containing grid-cell to grid-cell weights.
        exp_arg (np.ndarray): array of shape (N, N) representing the matrix being exponentiated.  
    """
    N = n_y * n_x
    exp_arg = np.zeros((N, N))

    R_beta = helpers.calc_rot_mat(beta)

    modulation_vec = alpha * R_beta @ velocity
    modulation_mat = np.zeros((2, N, N))
    modulation_mat[0,:,:] = modulation_vec[0]
    modulation_mat[1,:,:] = modulation_vec[1]
    
    cell_position_diffs_with_mod = cell_position_diffs + modulation_mat

    modded_mat = cell_position_diffs_with_mod.reshape(2, -1)
    exp_arg = calc_tri_dist_squared(modded_mat, use_GPU).reshape(N, N)
    
    exp_arg = -1*(exp_arg / np.power(sigma, 2))
    
    W = I * np.exp(exp_arg) - T
    return W, exp_arg

def update_function(A_t, W_t, A_l_t, W_l_t, tau, mean_val):
    """
    Updates the grid cell activity vector from one timestep to the next.
    
    Args:
        A_t (np.ndarray): array of shape (N,), where N = n_y * n_x, containing the activity of grid cells at a given timebin.
        W_t (np.ndarray): array of shape (N, N) containing the calculated grid-grid weight matrix at the associated timebin.
        A_l_t (np.ndarray): array of shape (n_landmarks, n_ln) containing the activity of the landmark cells.
        W_l_t (np.ndarray): array of shape (n_landmarks, n_ln, N) containing the weights of each landmark to each grid cell.
        tau (float): stabilization strength.
        mean_val (float): the mean of the transfer function in the previous timebin.
    Returns:
        A_t1 (np.ndarray): array of shape (N,), where N = n_y * n_x, containing the activity of grid cells at the next timebin.
        B_t1 (np.ndarray): array of shape (N,), where N = n_y * n_x, containing the linear transfer function at the next timebin.
    """
    B_t1 = A_t + W_t.T @ A_t + W_l_t.T @ A_l_t
    A_t1 = B_t1 + tau * B_t1 * (1 / mean_val - 1)
    A_t1 = helpers.nonnegative(A_t1)
    
    return A_t1, B_t1





import GPUtil

# Check if GPU exists
if len(GPUtil.getAvailable())>0:
    try:
        # Check if pytorch installed
        import torch

        # Check if GPU version of pytorch installed
        if torch.cuda.is_available():
            can_use_GPU = True
    except:
        can_use_GPU = False
else:
    can_use_GPU = False



import numpy as np
import sys
import os
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from ttgc.update import calc_tri_dist_squared
from ttgc import helpers

def calc_cell_positions(n_y, n_x):
    """
    Calculates the positions of each grid cell.
    
    Args:
        n_y (int): number of grid cells along the y-axis.
        n_x (int): number of grid cells along the x_axis.
    Returns:
        cell_positions (np.ndarray): array of shape (2, n_y, n_x) containing the y and x positions of each grid cell.
    """
    cell_positions = np.zeros((2, n_y, n_x))
    cell_positions[0, :, :] = np.tile(np.sqrt(3)/2 * (np.arange(1, n_y+1) - 0.5) / n_y, (n_x, 1)).T
    cell_positions[1, :, :] = np.tile((np.arange(1,n_x+1) - 0.5)/n_x, (n_y, 1))
  
    return cell_positions

def calc_cell_position_diffs(cell_positions):
    """
    Calculates the position differences between each pair of grid cells.
    
    Args:
        cell_positions (np.ndarray): array of shape (2, n_y, n_x) containing the y and x positions of each grid cell.
    Returns:
        cell_position_diffs (np.ndarray): array of shape (2, N, N), where N = n_y * n_x, containing the differences in positions between pairs of cells. 
                                          Element :,l,m represents the difference between the positions of cells c_l and c_m.
    """
    # if you want the compare cell c_l (at position i1j1) to cell c_m (at position i2j2), you get their 1d indices.
    _, n_y, n_x = cell_positions.shape
    N = n_y * n_x
    cell_position_diffs = np.zeros((2, N, N))

    for i in range(N):
        cell1_idx = helpers.one_to_two_index(i, n_y) # sequentially down the rows then left to right across columns
        cell1_pos = cell_positions[:, cell1_idx[0], cell1_idx[1]]
        
        for j in range(N):
            cell2_idx = helpers.one_to_two_index(j, n_y)
            cell2_pos = cell_positions[:, cell2_idx[0], cell2_idx[1]]
            cell_position_diffs[:, i, j] = cell1_pos - cell2_pos
            
    return cell_position_diffs

def initialize_network(n_y, n_x, seed=1):
    """
    Initializes the network of n_y * n_x grid cells 
    
    Args:
        n_y (int): number of grid cells along the y-axis.
        n_x (int): number of grid cells along the y-axis.
        seed (int, optional): for replication.
    Returns:
        init_state (np.ndarray): array of shape (N,) containing initialized network activity of all grid cells.
    """
    N = n_y * n_x

    rng = np.random.default_rng(seed=seed)
    init_state = rng.uniform(low=0, high=1/np.sqrt(N), size=(N,))
    
    return init_state

def initialize_W_l(n_landmarks, n_ln, l_pinning_n, N, l_use_nearby, cell_position_diffs, use_GPU, seed=1):
    """
    Generates the landmark cell to grid cell weight matrix. 
    
    Args:
        n_landmarks (int): number of landmarks.
        n_ln (int): number of landmark cells per landmark.
        l_pinning_n (int): number of grid cells each landmark cell projects to initially.
        l_use_nearby (bool): whether to use nearby grid cells in setting initial pinning phases.
        cell_position_diffs (np.ndarray) array of shape (2, N, N), where N = n_y * n_x, containing the differences in positions between pairs of cells. 
        use_GPU (bool): whether to use GPU.
        seed (int, optional): for replication.
    Returns:
        W_l0 (np.ndarray): array of shape (n_landmarks, n_ln, N) containing the initial weight from each landmark cell from each landmark to each grid cell
    """

    rng = np.random.default_rng(seed=seed)
    
    pinning_phases = rng.choice(N, n_landmarks*n_ln*l_pinning_n).reshape(n_landmarks, n_ln, l_pinning_n)
    W_l0 = np.zeros((n_landmarks, n_ln, N))
    # add input to specify landmark pinning phases later 
    # add another conditional here about hebbian plasticity term being 0. otherwise, W_l_input MUST be calculated each timestep
    for l in range(n_landmarks):
        for lc in range(n_ln):
            for li in range(l_pinning_n):
                if li==0:
                    this_cell = pinning_phases[l, lc, li]
                else:
                    if l_use_nearby:
                        cell0 = pinning_phases[l, lc, 0]
                        this_cell_dist_from_all = cell_position_diffs[:, cell0, :]
                        this_cell_tri_dist = calc_tri_dist_squared(this_cell_dist_from_all, use_GPU)
                        closest_cells = np.argsort(this_cell_tri_dist)
                        this_cell = closest_cells[li+1]
                    else:
                        this_cell = pinning_phases[l, lc, li]
    
                W_l0[l, lc, this_cell] = 1/(l_pinning_n*n_ln) # add input to specify landmark pinning weight
    return W_l0

def preallocate_sim(n_timebins, N):
    """
    Preallocate arrays for simulation
    
    Args:
        n_timebins (int): number of timebins.
        N (int): number of grid cells.
    Returns:
        A (np.ndarray): array of shape (n_timebins, N) containing the activities of grid cells over time.
        B (np.ndarray): array of shape (n_timebins, N) containing the linear transfer functions of grid cells over time.
        mean_vals (np.ndarray): array of shape (n_timebins) containing the mean of the linear transfer function over time.
        W (np.ndarray): array of shape (n_timebins, N, N) containing the grid-grid weights over time.
    """
    A = np.zeros((n_timebins, N))
    B = np.zeros((n_timebins, N))
    mean_vals = np.zeros((n_timebins,))
    W = np.zeros((n_timebins, N, N))
                 
    return A, B, mean_vals, W



from matplotlib import pyplot as plt
import sys
import os
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from ttgc import(helpers, initialize)

def plot_weight_mat(W, cell_idx, n_y, n_x, three_dim, cmap='magma'):
    '''
    W (np.ndarray): array of shape (N, N) containing the grid-grid weights at a particular time.
    cell_idx (int): reference cell to observe.
    n_y (int): number of grid cells along the y-axis.
    n_x (int): number of grid cells along the x_axis.
    three_dim (bool): whether to plot in 3d.
    cmap (str, optional): colormap.

    '''
    fig = plt.figure(figsize=(n_x/5, n_y/5))

    cell_positions = initialize.calc_cell_positions(n_y, n_x)
    cell_position_diffs = initialize.calc_cell_position_diffs(cell_positions)
    
    xs = cell_position_diffs[0, :, cell_idx]
    ys = cell_position_diffs[1, :, cell_idx]
    zs = W[:,cell_idx]

    if three_dim:
        ax = fig.add_subplot(projection='3d')
        ax.scatter(xs, ys, zs, c=zs, cmap=cmap)
    else:
        plt.scatter(xs, ys, c=zs, cmap=cmap)
        plt.xticks([])
        plt.yticks([])

    plt.scatter(0, 0, c='r',s=n_x*5,marker='*')



import sys
import os
import numpy as np
from matplotlib import pyplot as plt
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from ttgc import(helpers, initialize, update, landmark)

def run_sim2d(init_state, cell_position_diffs, 
              n_y, n_x, alpha, beta, I, sigma, T, tau, velocity, start_pos,
              l_pos, l_str, n_ln, l_pinning_n, l_use_nearby, l_lookahead,
              alpha_hebb,
              n_warmup_bins,
              B0, use_GPU, beta_noise=0, weight_noise=0, 
              W_input=None, W_l_input=None, seed=1, output_bins=10000):
    """
    Runs the full simulation given velocity inputs over time.
    
    Args:
        init_state (np.ndarray): array of shape (N,) containing initialized network activity of all grid cells.
        cell_position_diffs (np.ndarray) array of shape (2, N, N), where N = n_y * n_x, containing the differences in positions between pairs of cells. 
        n_y (int): number of grid cells along the y-axis.
        n_x (int): number of grid cells along the x_axis.
        alpha (float): gain.
        beta (float): bias.
        I (float): intensity parameter.
        sigma (float): Gaussian size.
        T (float): shift parameter.
        tau (float): stabilization strength.
        velocity (np.ndarray): array of shape (2,) representing the velocity at a timebin.
        start_pos (np.ndarray): array of shape (2,) representing the starting position of the agent.
        l_pos (np.ndarray): array of shape (n_landmarks, 2) representing the locations of landmarks in the environment. Input np.empty(0) if no landmarks.
        l_str (float): strength of landmark pinning.
        n_ln (int): number of landmark cells per landmark.
        l_pinning_n (int): number of grid cells each landmark cell projects to initially.
        l_use_nearby (bool): whether to use nearby grid cells in setting initial pinning phases.
        l_lookahead (float): the lookahead distance at which landmarks start recruiting their landmark cell(s) activity.
        alpha_hebb (float): Hebbian plasticity gain term.
        n_warmup_bins (int): number of timebins to use with 0 velocity to stabilize a bump of activity through attractor dynamics.
        B0 (np.ndarray): the linear transfer function at the 0th timebin. Input np.zeros(N) if no specific array to be used.
        use_GPU (bool): whether to use GPU.
        beta_noise (float, optional): Gaussian sigma for the amount of noise to be added to the rotation matrix per timebin. 
        weight_noise (float, optional): Gaussian sigma for the amount of noise to be added to grid-grid weight matrix per timebin.
        W_input (np.ndarray, optional): array of shape (n_timebins, N, N) representing the grid-grid weight matrix over time to be used. Useful for replicating simulations with equivalent velocity inputs.
        W_l_input (np.ndarray, optional): array of shape (n_timebins, n_landmarks, n_ln, N) representing the landmark cell to grid weight matrix over time to be used.
        seed (int, optional): for replication.
        output_bins (int, optional): number of timebins before next plot of population activity is rendered during simulation. Set to np.inf to suppress all plots.
    Returns:
        A (np.ndarray): array of shape (n_timebins, N) containing the activities of grid cells over time.
        B (np.ndarray): array of shape (n_timebins, N) containing the linear transfer functions of grid cells over time.
        W (np.ndarray): array of shape (n_timebins, N, N) containing the grid-grid weights over time.
        pos (np.ndarray): array of shape (n_timebins, 2) containing the position of the agent over time.
        A_l (np.ndarray): array of shape (n_timebins, n_landmarks, n_ln) containing the activities of the n_landmark cells over time.
        W_l (np.ndarray): array of shape (n_timebins, n_landmarks, n_ln, N) containing the weights between landmark cells and grid cells over time.
        betas_with_noise (np.ndarray): array of shape (n_timebins,) containing noisy beta values over time.
        weight_noise_vals (np.ndarray): array of shape (n_timebins, N, N) containing the grid-grid noise values added to weights over time.
    TODO:
        Incorporate hebbian plasticity mechanism; make sure to include alpha_decay term.
    """
    alpha_hebb = 0
    n_timebins = velocity.shape[0]
    N = init_state.shape[0]
    A, B, mean_vals, W = initialize.preallocate_sim(n_timebins, N)

    if W_input is not None:
        W = np.copy(W_input)
        betas_with_noise = np.zeros((n_timebins,))
        weight_noise_vals = np.zeros((n_timebins, N, N))
    else:
        W = np.zeros((n_timebins, N, N))
        rng = np.random.default_rng(seed=seed)
        betas_with_noise = beta+np.cumsum(beta_noise*rng.standard_normal(size=((n_timebins,))))
        weight_noise_vals = weight_noise*rng.standard_normal(size=(n_timebins, N, N))

    B[0,:] = B0
    mean_vals[0] = np.mean(B0)

    n_landmarks, l_pos = landmark.get_n_landmarks(l_pos)
    pos = np.cumsum(velocity, axis=0) + start_pos
    A_l = landmark.calc_A_l(pos, l_pos, n_ln, l_lookahead, l_str)

    if W_l_input is not None:
        W_l = W_l_input
    else:
        W_l = np.zeros((n_timebins, n_landmarks, n_ln, N))
        W_l[0] = initialize.initialize_W_l(n_landmarks, n_ln, l_pinning_n, N, l_use_nearby, cell_position_diffs, use_GPU, seed=seed)
        
        if alpha_hebb == 0:
            W_l[1:] = W_l[0]

    if n_warmup_bins > 0:
        A_warmup, B_warmup, mean_vals_warmup, W_warmup = initialize.preallocate_sim(n_warmup_bins, N)
        A_l_warmup = np.zeros((n_warmup_bins, n_landmarks, n_ln))
        W_l_warmup = np.zeros((n_warmup_bins, n_landmarks, n_ln, N))
        
        warmup_velocity = np.tile(np.zeros((1, 2)), (n_warmup_bins, 1))
        warmup_beta = beta * np.ones(n_warmup_bins)
        A_warmup[0, :] = init_state

        for t in np.arange(0, n_warmup_bins-1):
            if t == 0:
                mean_vals_warmup[t] = np.mean(A_warmup[t, :])
            else:
                mean_vals_warmup[t] = np.mean(B_warmup[t, :])
                            
            W_warmup[t, :, :], _, = update.calc_weight_mat(cell_position_diffs, n_y, n_x, alpha, warmup_beta[t], warmup_velocity[t], I, sigma, T, use_GPU)
            A_warmup[t+1, :], B_warmup[t+1, :] = update.update_function(A_warmup[t, :], W_warmup[t, :, :], A_l_warmup[t, :, :].reshape(-1), W_l_warmup[t, :, :, :].reshape(-1, N), 
                                                                        tau, mean_vals_warmup[t])

        init_state = A_warmup[-1, :]
        if not output_bins == np.inf:
            print('Warmup Complete')

    A[0, :] = init_state
    for t in np.arange(0, n_timebins-1):
        if mean_vals[0] == 0:
            mean_vals[t] = np.mean(A[t, :])
        else:
            mean_vals[t] = np.mean(B[t, :])

        # calculate weight matrix and add noise if not supplied
        if W_input is None:  
            if t==0:
                beta_t = beta
            else:
                beta_t = betas_with_noise[t-1]
            W[t, :, :], _ = update.calc_weight_mat(cell_position_diffs, n_y, n_x, alpha, beta_t, velocity[t], I, sigma, T, use_GPU)
            W[t, :, :] += weight_noise_vals[t, :, :] # add noise

        # update next timestep
        A[t+1, :], B[t+1, :] = update.update_function(A[t, :], W[t, :, :], A_l[t, :, :].reshape(-1), W_l[t, :, :, :].reshape(-1, N), tau, mean_vals[t])
        
        if (np.mod(t, output_bins)==0 and not output_bins==np.inf):
            if output_bins > 0:
                print(t)
                fig = plt.figure(figsize=(3,3))
                plt.imshow(A[t, :].reshape(n_x,n_y))
                plt.show()

    return A, B, W, pos, A_l, W_l, betas_with_noise, weight_noise_vals
  
def run_sim1d():
    A = 1
    return A




import numpy as np

def get_n_landmarks(l_pos):
    """
    Returns the number of landmarks based on l_pos and creates a pseudo-landmark if l_pos is empty.
    
    Args:
        l_pos (np.ndarray): array of shape (n_landmarks, 2) containing the x and y positions of each landmark.
    Returns:
        n_landmarks (int): number of landmarks present.
        l_pos (np.ndarray): array of shape (n_landmarks, 2) containing the x and y positions of each landmark. If input is empty, will return an array of shape (1, 2).
    """
    if l_pos.size > 0:
        n_landmarks = l_pos.size // 2 # two dimensions in l_pos
    else:
        # create a pseudo-landmark that is never accessible
        l_pos = np.expand_dims(np.array([-np.inf, -np.inf]), 0)
        n_landmarks = 1
    return n_landmarks, l_pos

def calc_A_l(pos, l_pos, n_ln, l_lookahead, l_str):
    """
    Calculates the activity of landmark cell(s) based on the agent's position over time and the landmark position(s).
    
    Args:
        pos (np.ndarray): array of shape (n_timebins, 2) representing the agent's 2D position over time.
        l_pos (np.ndarray): array of shape (n_landmarks, 2) representing the n_landmarks 2D position.
        n_ln (int): number of landmark cells per landmark.
        l_lookahead (float): the lookahead distance at which landmarks start recruiting their landmark cell(s) activity.
        l_str (float): strength of landmark pinning.
    Returns:
        A_l (np.ndarray): array of shape (n_timebins, n_landmarks, n_ln) containing the activity of each landmark cell's activity over time.
    """
    n_timebins = pos.shape[0]
    n_landmarks, l_pos = get_n_landmarks(l_pos)
    
    # create pseudo-landmark cell if n_ln set to anything less than 1
    if n_ln < 1:
        A_l = np.zeros((n_timebins, n_landmarks, 1))
    else:    
        pos_all = np.tile(np.expand_dims(pos, (1,2)), (1,n_landmarks,n_ln,1))
        l_pos_all = np.tile(np.expand_dims(l_pos, (0,2)), (n_timebins,1,n_ln,1))
        pos_from_l = pos_all - l_pos_all
        dist_from_l = np.linalg.norm(pos_from_l, axis=-1)

        if l_lookahead == 0:
            exp_factor = 0
        else:
            exp_factor = 2 / l_lookahead
            
        far_idxs = np.abs(dist_from_l) >= l_lookahead # the smaller this becomes, the larger the factor x needs to be below (x * (true_pos_from_l))
        dist_from_l[far_idxs] = -np.inf
        A_l = l_str * np.exp(-np.power(exp_factor*(dist_from_l),2)) 
        
    return A_l
    



import numpy as np

def one_to_two_index(idx, n_y):
    """
    Converts a single index to a two-element list representing the same index.
    
    Args:
        idx (int): single index.
        n_y (int): number of grid cells along the y-axis.
    Returns:
        two_index (list): row#, col#.
    """
    # assuming n_y is dimension 0 (row #), this will gives you a two_index of [row #, col #]
    two_index = []
    two_index.append(int(np.mod(idx, n_y)))
    two_index.append(int(idx - np.mod(idx, n_y))//n_y)
 
    return two_index

def relu(x, threshold=0):
    """
    Rectified linear unit.

    Args:
        x (float): input value.
        threshold (float, optional): rectified value. Defaults to 0.0.       
    Returns:
        float: the greater of x and threshold.
    """
    return np.maximum(x, threshold)
    
def calc_rot_mat(beta):
    """
    Calculates the 2x2 rotation matrix of angle beta (in radians).
    
    Args:
        beta (float): angle.
    Returns:
        R_beta (np.ndarray): array of shape (2, 2) representing the rotation matrix of angle beta.
    """
    R_beta = np.stack([[np.cos(beta), -np.sin(beta)], [np.sin(beta), np.cos(beta)]])
                      
    return R_beta

def nonnegative(x):
    """
    Nonnegativity function 
    
    Args:
        x (np.ndarray or float): input.
    Returns:
        nonneg_x (np.ndarray or float): nonnegative version of x.
    """
    nonneg_x = x.clip(min=0)
    return nonneg_x

def str_to_float(arr):
    """
    Convert an array of string values to float values. np.nan values filled where errors occur.
    
    Args:
        arr (np.ndarray): input array with string values. 
    Returns:
        np.ndarray: array with float values.
    """
    convert_arr = []
    
    for s in arr.ravel():    
        try:
            value = float(s)
        except ValueError:
            value = np.nan

        convert_arr.append(value)

    return np.array(convert_arr, dtype=object).reshape(arr.shape)

def calc_2d_maps(A, pos, n_bins):
    """
    Calculates an estimate of 2D rate maps
    
    Args:
        A (np.ndarray): array of shape (n_timebins, N) containing the activities of N cells over time.
        pos (np.ndarray): array of shape (n_timebins, 2) containing the position of the agent over time.
        n_bins (int): number of spatial bins to discretize the environment.
    Returns:
        mean_maps (np.ndarray): array of shape (N, n_bins, n_bins) containing the mean firing activity of each cell.
    """
    n_timebins, N = np.shape(A)
    H = np.digitize(pos, np.linspace(0, np.max(pos), n_bins+1), right=True)
    
    wherebins = []
    for xi in range(n_bins):
        tempx = np.where(H[:, 0] == xi+1)[0]
        
        for yi in range(n_bins):
            tempy = np.where(H[:,1]==yi+1)[0]
    
            wherebins.append(np.intersect1d(tempx,tempy))
    
    mean_maps = np.zeros((N, n_bins, n_bins))
    
    for xi in range(n_bins):
        
        for yi in range(n_bins):
            these_bins = wherebins[xi+yi*n_bins]
            
            if these_bins.size>0:
                mean_maps[:, xi, yi] = np.mean(A[these_bins, :], axis=0)
                
    return mean_maps